# #####################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2025                                            #
# #####################################################################

###
# The main function to run the any analysis for fiber
# Working:
# - Get data set work on one dataset at a time, so require file not folder
# - flags to run globla analysis
#     - Lenght
#     - Curv/tourrositi
#     - fiber classification (require chromosome and poles)
#     - Global interaction map

# The rest need napari ...
###
